
/*
 * dataman.c - data manager module.
 * $Id: dataman.c,v 1.7 1998/06/01 00:54:59 ademko Exp $
 *
 * Copyright (C) 1998 Alex Demko <ademko@mbnet.mb.ca>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* for routines in live parser */
#include <ctype.h>

/* for dir reading routines */
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>

#include <dataman.h>
#include <defines.h>

/* local static vars */
const char *wmconfigDir;

/*
 * parse_line()
 *
 * Parses one line in a wmconfig file (taken from wmconfig's sources)
 * The routine is (C)1997 Red Hat Software, Inc. (used under the GNU GPL)
 *
 */

/* marcos used in parline */
#define SKIP_SPACES(c)  while (*c && isspace(*c)) c++ 
#define SKIP_CHARS(c)   while (*c && !isspace(*c)) c++ 
#define SKIP_TO(c, x)   while (*c && (*c != x)) c++
#define FINISH(code)    { *package=item1; *tag=item2; *string=item3; return(code); }

static int parse_line(char *buffer, char **package, char **tag, char **string)
{
    char *item1 = NULL;
    char *item2 = NULL;
    char *item3 = NULL;
    char *c;
    
    /* verify the line */
    c = buffer;
    SKIP_SPACES(c);

    /* skip comments */
    if (*c && (*c == '#'))
    FINISH(-1);
    
    /* okay, not a comment */
    item1 = c; /* the program goes first */
    SKIP_CHARS(c);
    if (*c)
    *c++ = '\0'; /* that was the program */
    else 
    FINISH(1);
    
    SKIP_SPACES(c);
    if (*c)
    item2 = c; /* try to read the tag */
    else
    FINISH(1);

    SKIP_CHARS(c);
    if (*c)
    *c++ = '\0'; /* that was the tag */
    else
    FINISH(2);

    SKIP_SPACES(c);
    if (*c)
    item3 = c;
    else
    FINISH(2);

    if (*c == '"') {
    item3++; c++;
    SKIP_TO(c, '"');
    if (*c == '"')
        *c='\0';
    } else  
    SKIP_CHARS(c);

    if (*c)
    *c++ = '\0';

    FINISH(0);
}    

/*
 * newApp()
 *
 * creates a new, empty app
 */
app *newApp(void) {
    app *n;

    /* allocate the new app */
    n = g_malloc0(sizeof(app));
    if (!n)
       error_bomb("newApp(): out of memory")

    return n;
}

/*
 * freeApp()
 *
 * frees an app from memory and all its fields
 * (except 'next')
 */
void freeApp(app *a) {
    /* make sure we have something */
    if (!a)
       return;

    /* nuke all the non null fields */
    sfree(a->id)
    sfree(a->name)
    sfree(a->icon)
    sfree(a->mini_icon)
    sfree(a->description)
    sfree(a->Xresource)
    sfree(a->exec)
    sfree(a->group)
    sfree(a->copy)
    sfree(a->filename)

    g_free(a);
}

/*
 * debugDumpAppList()
 *
 */
static void debugDumpAppListInterator(app * data, gpointer user_data) {
    g_print("id = %s\n",data->id);
}
void debugDumpAppList(GList *list) {
    if (!list) {
        g_warning("debugDumpAppList() was sent a NULL list");
        return;
    }

    g_list_foreach(list,(GFunc)debugDumpAppListInterator,NULL);
}

/*
 * killList()
 *
 * kills the given list
 */
static void freeListInterator(app * data, gpointer user_data) {
    freeApp(data);
}
void freeList(GList *list) {
    g_list_foreach(list,(GFunc)freeListInterator,NULL);
    g_list_free(list);
}

/*
 * processAppLine()
 *
 * process a wmconfig line
 */
/* simple string comparison macro */
#define STREQ(s1,s2) (strcasecmp((s1),(s2)) == 0)
static void processAppLine(app *App, char *line) {
    char *name, *tag, *value;
    int retc;

    /* parse this line */
    if ( (retc = parse_line(line,&name,&tag,&value)) ) {
       g_warning("error parsing:  \"%s\"\n",line);
       return;
    }

    if (!App->id)
        App->id = sstrdup(name);

    /* set the right tag in the node */
    if (STREQ(tag,"name"))
       App->name = sstrdup(value);
    else if (STREQ(tag,"icon"))
       App->icon = sstrdup(value);
    else if (STREQ(tag,"description"))
       App->description = sstrdup(value);
    else if (STREQ(tag,"mini-icon"))
       App->mini_icon = sstrdup(value);
    else if (STREQ(tag,"exec"))
       App->exec = sstrdup(value);
    else if (STREQ(tag,"Xresource"))
       App->Xresource = sstrdup(value);
    else if (STREQ(tag,"group"))
       App->group = sstrdup(value);
    else if (STREQ(tag,"copy"))
       App->copy = sstrdup(value);
}

/*
 * appendApp()
 *
 * add a new app to the list
 */
void appendApp(GList *list, app *App) {
    list = g_list_append(list,App);
}
 
/*
 * addAppFile
 *
 * adds an app file to the given list
 */
static GList * addAppFile(GList *list, const char *fullname,
        const char *filename) {
    FILE *in;
    app *App;
    char buf[MAX_S];

    in = fopen(fullname,"r");
    if (!in) {
        g_warning("can't open wmconfig file: %s\n",fullname);
        return list;
    }

    App = newApp();
    App->filename = sstrdup(filename);

    /* process each line of the file */
    while ( fgets(buf,MAX_S,in) ) {
       /* if it's long enough, chop of the newline
          and feed it into the line processor */
       if (strlen(buf)<4)
          continue; /* to short */
       buf[strlen(buf)-1] = '\0';

       /* process the line */
       processAppLine(App,buf);
    }

    list = g_list_append(list,App);

    fclose(in);

    return list;
}

/*
 * setWMCDir()
 *
 * set main wmconfig dir
 */
void setWMCDir(const char *dirname) {
    wmconfigDir = dirname;
}

/*
 * addWMCDir()
 *
 * adds a directoy full of wmconfig files to the given
 * root list
 */
#define minstring(num) { \
    if (bufsize<num) { \
    buf = g_realloc(buf,num); \
    bufsize = num; \
    } }
GList * addWMCDir(GList *list, const char *dirname) {
    DIR *dir;
    struct dirent *ent;
    struct stat st;
    char *buf; /* dymanic string */
    int bufsize;

    /* make sure we have a menu to work with */
    /*if (!list)
       error_bomb("root list must be created (!null)");*/

    /* allocate an initial buffer string */
    buf = g_malloc(MAX_S); /* if more is needed, it will be expanded */
    bufsize = MAX_S;

    /* open the given dir */
    if ( !(dir=opendir(dirname)) )
       return list; /* error - bail */

    /* interate through all the files in to directory */
    while ( (ent = readdir(dir)) ) {
       /* make sure our buffer is big enough and build the filename */
       minstring(strlen(dirname) + strlen(ent->d_name) + 5)
       sprintf(buf,"%s/%s",dirname,ent->d_name);

       /* get the file info */
       stat(buf,&st);

       /* if it's a file, update the work buffer and operate on the file */
       if (S_ISREG(st.st_mode))
          list = addAppFile(list,buf,ent->d_name);
    }

    /* close the dir and return successfully */
    closedir(dir);

    return list;
}

/*
 * deleteAppFile()
 *
 * deletes just the physical file
 */
void deleteAppFile(const char *filename) {\
    if (filename) {
        char *fullname;
        
        /* build the filename and nuke the file */
        fullname = g_malloc(strlen(wmconfigDir)+strlen(filename)+5);

        sprintf(fullname,"%s/%s",wmconfigDir,filename);
        if (unlink(fullname))
            perror(fullname);

        g_free(fullname);
    }
}

/*
 * deleteApp()
 *
 * deletes the app from the list and file
 */
void deleteApp(GList *list, app *App) {
    /* remove the item from the list */
    g_list_remove(list,App);
    /* deletes the file */
    if (App->filename)
        deleteAppFile(App->filename);
    /* free the app */
    freeApp(App);
}

/*
 * updateApp()
 *
 * write the app to the physical file
 */
#define APPFILE_LINE(file,tag,name,val) if (val && (strlen(val)>0)) \
        fprintf((file),"%s %s \"%s\"\n",(tag),(name),(val))
void updateApp(app *App) {
    FILE *out;
    char *fullname;

    /* check for decent parameters */
    if (!App || !App->filename) {
        g_warning("updateApp() null App or App->filename sent\n");
        return;
    }

    /* build the full file name */
    fullname = g_malloc(strlen(wmconfigDir)+strlen(App->filename)+5);
    sprintf(fullname,"%s/%s",wmconfigDir,App->filename);

    /* open the file for output */
    out = fopen(fullname,"w");
    if (!out) {
        g_warning("updateApp() error opening \"%s\" for writing\n",fullname);
        g_free(fullname);
        return;
    }

    /* write all the data */
    APPFILE_LINE(out,App->id,"name",App->name);
    APPFILE_LINE(out,App->id,"icon",App->icon);
    APPFILE_LINE(out,App->id,"mini-icon",App->mini_icon);
    APPFILE_LINE(out,App->id,"description",App->description);
    APPFILE_LINE(out,App->id,"Xresource",App->Xresource);
    APPFILE_LINE(out,App->id,"exec",App->exec);
    APPFILE_LINE(out,App->id,"group",App->group);
    APPFILE_LINE(out,App->id,"copy",App->copy);

    fclose(out);
    g_free(fullname);
}

